/* BlinkenLib
   Copyright 2004-2016 Stefan Schuermans <stefan@schuermans.info>
   Copyleft GNU public license - http://www.gnu.org/copyleft/gpl.html
   a blinkenarea.org project */

#include <gif_lib.h>

#include <BlinkenLib/config.h>
#include <BlinkenLib/BlinkenConstants.h>
#include <BlinkenLib/BlinkenFrame.h>
#include <BlinkenLib/BlinkenMovie.h>
#include <BlinkenLib/BlinkenGif.h>

// get color from GIF color map
static void BlinkenGifGetColor(ColorMapObject *pMap, int idx, int transparent,
                               unsigned char *r, unsigned char *g,
                               unsigned char *b, unsigned char *t)
{
  if (transparent >= 0 && idx == transparent) {
    *t = 1;
  } else {
    *t = 0;
  }
  if (pMap != NULL) {
    if (idx >= 0 && idx < pMap->ColorCount) {
      *r = pMap->Colors[idx].Red;
      *g = pMap->Colors[idx].Green;
      *b = pMap->Colors[idx].Blue;
    } else {
      *r = 0;
      *g = 0;
      *b = 0;
      *t = 1;
    }
  } else {
    *r = 0;
    *g = 0;
    *b = 0;
    *t = 1;
  }
}

// set pixel to GIF color
static void BlinkenGifSetPixel(ColorMapObject *pMap, int idx, int transparent,
                               stBlinkenFrame *pFrame, int y, int x)
{
  unsigned char r, g, b, t;

  if (pFrame != NULL) {
    BlinkenGifGetColor(pMap, idx, transparent, &r, &g, &b, &t);
    if (! t) {
      BlinkenFrameSetPixel(pFrame, y, x, 0, r);
      BlinkenFrameSetPixel(pFrame, y, x, 1, g);
      BlinkenFrameSetPixel(pFrame, y, x, 2, b);
    }
  }
}

// load a GIF file as BlinkenMovie
stBlinkenMovie *BlinkenGifLoad(const char *pFilename)
{
  GifFileType *gif;
  int height, width, frameCnt, frameIdx, i, y, x,
      bg, disposal, delay, transp, idx;
  ColorMapObject *pGlobalMap, *pMap;
  SavedImage *pImg;
  GifImageDesc *pDesc;
  ExtensionBlock *pEx;
  stBlinkenMovie *pMovie;
  stBlinkenFrame *pBack, *pFrame;

  if (pFilename == NULL)
    return NULL;

  // open GIF file for decoding
  gif = DGifOpenFileName(pFilename);
  if (gif == NULL) {
    return NULL;
  }

  // read GIF file
  if (DGifSlurp(gif) != GIF_OK) {
    DGifCloseFile(gif);
    return NULL;
  }
  height = gif->SHeight;
  width = gif->SWidth;
  frameCnt = gif->ImageCount;
  pGlobalMap = gif->SColorMap;
  bg = gif->SBackGroundColor;

  // create movie
  pMovie = BlinkenMovieNew(height, width, 3, 255);
  if (! pMovie) {
    DGifCloseFile(gif);
    return NULL;
  }

  // create background frame and set it to background color
  pBack = BlinkenFrameNew(height, width, 3, 255, 1);
  if (! pBack) {
    BlinkenMovieFree(pMovie);
    DGifCloseFile(gif);
    return NULL;
  }
  for (y = 0; y < height; ++y) {
    for (x = 0; x < width; ++x) {
      BlinkenGifSetPixel(pGlobalMap, bg, -1, pBack, y, x);
    }
  }

  // process all frames
  for (frameIdx = 0; frameIdx < frameCnt; ++frameIdx) {
    pImg = &gif->SavedImages[frameIdx];
    pDesc = &pImg->ImageDesc;

    // get local color map
    pMap = pDesc->ColorMap != NULL ? pDesc->ColorMap : pGlobalMap;

    // get disposal mode, delay and transparent color
    disposal = 0;
    delay = 100;
    transp = -1;
    for (i = 0; i < pImg->ExtensionBlockCount; ++i) {
      pEx = &pImg->ExtensionBlocks[i];
      if (pEx->Function == GRAPHICS_EXT_FUNC_CODE && pEx->ByteCount == 4) {
        disposal = pEx->Bytes[0] >> 2 & 0x07;
        delay = (unsigned int)(unsigned char)pEx->Bytes[1] |
                (unsigned int)(unsigned char)pEx->Bytes[2] << 8;
        transp = pEx->Bytes[0] & 0x01 ? (int)(unsigned char)pEx->Bytes[3] : -1;
      }
    }

    // draw new frame based on background frame, maybe update background frame
    if (delay > 0) {
      pFrame = BlinkenFrameClone(pBack);
      BlinkenFrameSetDuration(pFrame, delay * 10); // GIF delay is in 10ms
    } else {
      // frames with no duration -> only effect on background
      pFrame = NULL;
    }
    // FIXME: pDesc->Interlace > 0
    for (y = pDesc->Top, i = 0; y < pDesc->Top + pDesc->Height; ++y) {
      for (x = pDesc->Left; x < pDesc->Left + pDesc->Width; ++x, ++i) {
        idx = pImg->RasterBits[i];
        switch (disposal) {
          // undefined
          case 0:
          case 4:
          case 5:
          case 6:
          case 7:
          // do not dispose -> draw to frame and update background
          case 1:
            BlinkenGifSetPixel(pMap, idx, transp, pFrame, y, x);
            BlinkenGifSetPixel(pMap, idx, transp, pBack, y, x);
            break;
          // restore background -> draw to frame and reset background
          case 2:
            BlinkenGifSetPixel(pMap, idx, transp, pFrame, y, x);
            BlinkenGifSetPixel(pMap, bg, transp, pBack, y, x);
            break;
          // restore previous -> draw to frame only
          case 3:
            BlinkenGifSetPixel(pMap, idx, transp, pFrame, y, x);
            break;
        }
      }
    }

    // append frame to movie
    if (pFrame) {
      if (BlinkenMovieAppendFrame(pMovie, pFrame) != 0) {
        BlinkenFrameFree(pFrame);
      }
    }

  } // for frameIdx

  // delete background frame
  BlinkenFrameFree(pBack);

  // close GIF file
  DGifCloseFile(gif);

  return pMovie;
}

